#include <iostream> 
#include <fstream>
#include <utility>
#include <string>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/random.hpp>
#include <boost/random.hpp>
#include <boost/graph/astar_search.hpp>
#include <vector>
#include <ctime>
#include <map>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>
#include <libpq-fe.h>
#ifndef ROADNETWORK_H
#define ROADNETWORK_H
using namespace std;

// NETWORK STRUCTS AND FUNCTIONS
float ALPHA = 0.15;
float BETA = 4.0;
extern bool DEBUG;

typedef float cost;

struct Node
{
	int id;
    double lon;
    double lat;
};

struct Edge
{
    int id;
    unsigned int flow;
    int capacity;
    int speed;
    double travel_distance;
    double free_travel_time;
    double travel_time;
};

typedef boost::adjacency_list<boost::vecS, 
                            boost::vecS, 
                            boost::directedS,
                            Node,
                            Edge> RoadNetwork;

typedef boost::graph_traits<RoadNetwork>::vertex_descriptor vertex_descriptor;
typedef boost::graph_traits<RoadNetwork>::edge_descriptor edge_descriptor;

// A hash function for edges.
struct edge_hash:unary_function<edge_descriptor, size_t> 
{
    edge_hash(RoadNetwork const& g):g(g){}

    size_t operator()(edge_descriptor const& e) const {
    size_t seed = 0;
    boost::hash_combine(seed, source(e,g));
    boost::hash_combine(seed, target(e,g));
    return seed;
    }
    RoadNetwork const& g;
};

typedef boost::property_map<RoadNetwork, boost::vertex_bundle_t>::type NodeMap;
typedef map<edge_descriptor, unsigned int > EdgeMap;
typedef map<edge_descriptor, map<int,pair<int,int> > > EdgeTractMap;

// euclidean distance heuristic
template <class Graph, class CostType, class LocMap>
class distance_heuristic : public boost::astar_heuristic<Graph, CostType>
{
    public:
        typedef typename boost::graph_traits<Graph>::vertex_descriptor Vertex;
        distance_heuristic(LocMap l, Vertex goal) : m_location(l), m_goal(goal) {}
        CostType operator()(Vertex u)
        {
            CostType dx = m_location[m_goal].lon - m_location[u].lon;
            CostType dy = m_location[m_goal].lat - m_location[u].lat;
            return ::sqrt(dx * dx + dy * dy);
        }
    private:
        LocMap m_location;
        Vertex m_goal;
};


struct found_goal {}; // exception for termination

// visitor that terminates when we find the goal
template <class Vertex>
class astar_goal_visitor : public boost::default_astar_visitor
{
    public:
        astar_goal_visitor(Vertex goal) : m_goal(goal) {}
        template <class Graph>
        void examine_vertex(Vertex u, Graph& g) {
            if(u == m_goal)
                throw found_goal();
        }
    private:
        Vertex m_goal;
};

// FUNCTIONS
static void exit_nicely(PGconn *conn)
{
    PQfinish(conn);
    exit(1);
}

RoadNetwork createNetworkFromFile(const string& fname_node, const string& fname_edge)
{
    cout << "Creating road network from file." << endl;
    RoadNetwork g;
    int id;
    double lon, lat;
    vertex_descriptor u;
    Node intr;
    ifstream fid1 (fname_node.c_str());
    while (fid1 >> id)
    {
        fid1 >> lon;
        fid1 >> lat;
        intr.lon = lon;
        intr.lat = lat;
        intr.id = id-1;
        u = boost::add_vertex(intr, g);
    }
    fid1.close();

    int s, t, cap, speed;
    double travel_time;
    if(DEBUG) cout << fname_edge << endl;

    ifstream fid (fname_edge.c_str());
    Edge rd;
    string head;
    getline(fid, head);
    while(fid>>id)
    {
        fid>>s;
        fid>>t;
        fid>>cap;
        fid>>speed;
        fid>>travel_time;
        
        // set the edge properties and add the edge
        rd.id = id-1;
        rd.flow = 0;
        rd.capacity = cap;
        rd.speed = speed;
        rd.travel_distance = 0.0;
        rd.free_travel_time = travel_time;
        rd.travel_time = travel_time;
        boost::add_edge(s-1,t-1,rd,g);
    }
    fid.close();
    return g;
}

RoadNetwork createNetworkFromDatabase(const std::string& table)
{
    cout << "Creating road network from database..." << endl;
    
    RoadNetwork g;
    PGconn *conn;
    PGresult *res;
    string query;
    Node intersection;
    Edge rd;
    int s, t;
    
    // connect to the database and make sure everything is ok
    conn = PQconnectdb("dbname=calles host=ec2-54-200-71-200.us-west-2.compute.amazonaws.com port=5432 user=jameson password=jameson");
    if (PQstatus(conn) != CONNECTION_OK)
    {
        fprintf(stderr, "Connection to database failed: %s",
                PQerrorMessage(conn));
        exit_nicely(conn);
    }
    // build and execute query for nodes
    query = "SELECT id, ST_X(the_geom), ST_Y(the_geom) FROM ";
    query.append(table);
    query.append("_vertices_pgr;");

    res = PQexec(conn, query.c_str());
    // make sure we get a result
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
        fprintf(stderr, "SELECT failed: %s", PQerrorMessage(conn));
        PQclear(res);
        exit_nicely(conn);
    }
   
    for( int i=0; i < PQntuples(res); i++)
    {
        intersection.id = atoi(PQgetvalue(res, i, 0))-1;
        intersection.lon = atof(PQgetvalue(res, i, 1));
        intersection.lat = atof(PQgetvalue(res, i, 2));
        if(DEBUG) cout << intersection.id <<','<< intersection.lon <<','<< intersection.lat << endl;
        boost::add_vertex(intersection, g);
    }
  
    // build and execute query for edges
    query = "SELECT gid, source, target, capacity, cost_time, speed_kmh, len_km FROM ";
    query.append(table);
    query.append(" ORDER BY gid ASC;");

    res = PQexec(conn, query.c_str());
    // make sure we get a result
    if (PQresultStatus(res) != PGRES_TUPLES_OK)
    {
        fprintf(stderr, "SELECT failed: %s", PQerrorMessage(conn));
        PQclear(res);
        exit_nicely(conn);
    }
    // read the tuples into edges of the BGL graph
    for( int i=0; i < PQntuples(res); i++)
    {
        rd.id = atoi(PQgetvalue(res, i, 0))-1;
        s = atoi(PQgetvalue(res, i, 1))-1;
        t = atoi(PQgetvalue(res, i, 2))-1;
        rd.capacity = atoi(PQgetvalue(res, i, 3));
        rd.free_travel_time = atof(PQgetvalue(res, i, 4));
        rd.travel_time = rd.free_travel_time;
        rd.speed = atoi(PQgetvalue(res, i, 5));
        rd.flow = 0;
        rd.travel_distance = atof(PQgetvalue(res, i, 6));
        if(DEBUG) cout << rd.id << ',' << s <<','<< t <<','<< rd.capacity <<','<< rd.travel_time << ',' << rd.flow <<endl;
        boost::add_edge(s,t,rd,g);
    }
    PQclear(res);

    PQfinish(conn);
    return g;
}

void updateDatabaseNetwork(const std::string& table, RoadNetwork* g)
{
    PGconn *conn;
    PGresult *res;
    conn = PQconnectdb("dbname=calles host=ec2-54-200-71-200.us-west-2.compute.amazonaws.com port=5432 user=jameson password=jameson");
    if (PQstatus(conn) != CONNECTION_OK)
    {
        fprintf(stderr, "Connection to database failed: %s",
                PQerrorMessage(conn));
        exit_nicely(conn);
    }

    /* Start a transaction block */
    res = PQexec(conn, "BEGIN");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        fprintf(stderr, "BEGIN command failed: %s", PQerrorMessage(conn));
        PQclear(res);
        exit_nicely(conn);
    }
    string longquery = "";
    boost::graph_traits<RoadNetwork>::edge_iterator ei, ei_end;
    int count = 0;
    for (boost::tie(ei, ei_end) = boost::edges(*g); ei != ei_end; ++ei)
    {
        longquery.append("UPDATE ");
        longquery.append(table);
        longquery.append(" SET voc=");
        longquery.append(static_cast<ostringstream*>( &(ostringstream() << (float)(*g)[*ei].flow/(float)(*g)[*ei].capacity) )->str());
        longquery.append(" WHERE gid=");
        longquery.append(static_cast<ostringstream*>( &(ostringstream() << (*g)[*ei].id+1) )->str());
        //longquery.append(" AND target=");
        //longquery.append(static_cast<ostringstream*>( &(ostringstream() << boost::target(*ei,*g)+1) )->str());
        longquery.append("; ");
        count ++;
        // update in batches of 1000 over slow connections
        if (count > 1000)
        {
            res = PQexec(conn, longquery.c_str());
            if (PQresultStatus(res) != PGRES_COMMAND_OK)
            {
                fprintf(stderr, "BEGIN command failed: %s", PQerrorMessage(conn));
                PQclear(res);
                exit_nicely(conn);
            }
            count = 0;
            longquery = "";
        }
    }
    res = PQexec(conn, longquery.c_str());
    res = PQexec(conn, "COMMIT");
    PQclear(res);
    PQfinish(conn);
}

void updateDatabaseUsage(const std::string& table, RoadNetwork* g, EdgeTractMap* tractMap)
{
    PGconn *conn;
    PGresult *res;
    conn = PQconnectdb("dbname=calles host=ec2-54-200-71-200.us-west-2.compute.amazonaws.com port=5432 user=jameson password=jameson");
    if (PQstatus(conn) != CONNECTION_OK)
    {
        fprintf(stderr, "Connection to database failed: %s",
                PQerrorMessage(conn));
        exit_nicely(conn);
    }

    /* Start a transaction block */
    res = PQexec(conn, "BEGIN");
    if (PQresultStatus(res) != PGRES_COMMAND_OK)
    {
        fprintf(stderr, "BEGIN command failed: %s", PQerrorMessage(conn));
        PQclear(res);
        exit_nicely(conn);
    }
    string longquery = "TRUNCATE ";
    longquery.append(table);
    longquery.append(";");
    PQexec(conn, longquery.c_str());
    
    longquery = "INSERT INTO ";
    longquery.append(table);
    longquery.append(" (gid, fips, flow_to, flow_from) VALUES ");
    boost::graph_traits<RoadNetwork>::edge_iterator ei, ei_end;
    typedef map<int, pair<int,int> >::iterator it_type;
    int count = 0;
    for (boost::tie(ei, ei_end) = boost::edges(*g); ei != ei_end; ++ei)
    {
        int gid = (*g)[*ei].id;
        for(it_type it = (*tractMap)[*ei].begin(); it != (*tractMap)[*ei].end(); it++)
        {
            if (count != 0) longquery.append(" , ");
            int fips = it->first;
            pair<int,int> flow = it->second;
            longquery.append(" (");
            longquery.append(static_cast<ostringstream*>( &(ostringstream() << gid+1) )->str());
            longquery.append(",");
            longquery.append(static_cast<ostringstream*>( &(ostringstream() << fips) )->str());
            longquery.append(",");
            longquery.append(static_cast<ostringstream*>( &(ostringstream() << flow.first) )->str());
            longquery.append(",");
            longquery.append(static_cast<ostringstream*>( &(ostringstream() << flow.second) )->str());
            longquery.append(")");
            count ++;
        }
        // update in batches of 2000 over slow connections
        if (count > 2000)
        {
            longquery.append(";");
            res = PQexec(conn, longquery.c_str());
            if (PQresultStatus(res) != PGRES_COMMAND_OK)
            {
                fprintf(stderr, "BEGIN command failed: %s", PQerrorMessage(conn));
                PQclear(res);
                exit_nicely(conn);
            }
            count = 0;
            longquery = "INSERT INTO ";
            longquery.append(table);
            longquery.append(" (gid, fips, flow_to, flow_from) VALUES ");
        }
    }
    longquery.append(";");
    res = PQexec(conn, longquery.c_str());
    res = PQexec(conn, "COMMIT");
    PQclear(res);
    PQfinish(conn);
}

void updateTravelTimes(RoadNetwork* g, EdgeMap* volume)
{
    float new_travel_time;
    float cap;
    float new_vol;
    boost::graph_traits<RoadNetwork>::edge_iterator ei, ei_end;
    for (boost::tie(ei, ei_end) = boost::edges(*g); ei != ei_end; ++ei)
    {
        cap = (float)(*g)[*ei].capacity;
        new_vol = (float)(*g)[*ei].flow + (float)(*volume)[*ei];
        new_travel_time = (*g)[*ei].free_travel_time*(1.0 + ALPHA*pow(new_vol/cap,BETA));
        (*g)[*ei].travel_time = new_travel_time;
        (*g)[*ei].flow = new_vol;
        if (DEBUG)
            cout << "Old Time, New Time: " << (*g)[*ei].free_travel_time << "," << new_travel_time << endl;
    }
}

void printEdges(const string& wfname, RoadNetwork* g, EdgeTractMap* tractMap)
{
    ofstream wfid (wfname.c_str());
    wfid << "gid s t vol cap voc k" << endl;
    boost::graph_traits<RoadNetwork>::edge_iterator ei, ei_end;
    for (boost::tie(ei, ei_end) = boost::edges(*g); ei != ei_end; ++ei)
    {
            wfid << (*g)[*ei].id+1 << ' ';
            //wfid << (*g)[boost::source(*ei,*g)].lat << ' ';
            //wfid << (*g)[boost::source(*ei,*g)].lon << ' ';
            //wfid << (*g)[boost::target(*ei,*g)].lat << ' ';
            //wfid << (*g)[boost::target(*ei,*g)].lon << ' ';
            wfid << boost::source(*ei,*g) + 1 << ' ';
            wfid << boost::target(*ei,*g) + 1 << ' ';
            wfid << (*g)[*ei].flow << ' '; 
            wfid << (*g)[*ei].capacity << ' '; 
            wfid << std::setprecision(6) << (float)(*g)[*ei].flow/(float)(*g)[*ei].capacity << ' ';
            wfid << (*tractMap)[*ei].size();
            wfid << endl;
    }
    wfid.close();
}
#endif

